mod extract;

use std::collections::HashMap;
use std::net::SocketAddr;
use std::path::{self, PathBuf};
use std::sync::Arc;

use crate::certs::{load_cert, read_certs, read_pubkey, store_cert};
use anyhow::Context;
use axum::body::{self};
use axum::extract::{Path, State};
use axum::routing::{post, put};
use axum::{http::StatusCode, response::IntoResponse, Router};
use axum_extra::routing::{
    RouterExt, // for `Router::typed_*`
    TypedPath,
};
use clap::Parser;
use serde::Deserialize;
use ssh_key::{Certificate, PublicKey};
use tokio::sync::Mutex;
use tower::ServiceBuilder;
use tower_http::ServiceBuilderExt;

use self::extract::CertificateBody;

#[derive(Parser)]
pub struct ApiArgs {
    #[clap(short = 'a', long = "address")]
    address: SocketAddr,
    #[clap(short = 'c', long = "cert-store")]
    cert_dir: PathBuf,
    /// CA public key
    #[clap(long = "ca")]
    ca: PathBuf,
}

impl Default for ApiArgs {
    fn default() -> Self {
        Self {
            address: SocketAddr::from(([127, 0, 0, 1], 3000)),
            cert_dir: "certs".into(),
            ca: "certs/ca.pub".into(),
        }
    }
}

#[derive(Debug, Clone)]
struct ApiState {
    certs: Arc<Mutex<HashMap<String, Certificate>>>,
    cert_dir: PathBuf,
    ca: PublicKey,
}

impl ApiState {
    async fn new(
        cert_dir: impl AsRef<path::Path>,
        ca_file: impl AsRef<path::Path>,
    ) -> anyhow::Result<Self> {
        let ca = read_pubkey(ca_file.as_ref()).await?;
        let certs = read_certs(&ca, cert_dir.as_ref()).await?;
        Ok(Self {
            certs: Arc::new(Mutex::new(
                certs
                    .into_iter()
                    .map(|cert| (cert.key_id().to_string(), cert))
                    .collect(),
            )),
            cert_dir: cert_dir.as_ref().into(),
            ca,
        })
    }
}

pub async fn run(
    ApiArgs {
        address,
        cert_dir,
        ca,
    }: ApiArgs,
) -> anyhow::Result<()> {
    let state = ApiState::new(&cert_dir, &ca).await?;

    #[cfg(feature = "reload")]
    {
        let state = state.clone();

        tokio::spawn(async move {
            loop {
                tokio::time::sleep(Duration::from_secs(30)).await;
                if let Ok(certs) = read_certs(&state.ca, &state.cert_dir).await {
                    *state.certs.lock().await = certs
                        .into_iter()
                        .map(|cert| (cert.key_id().to_string(), cert))
                        .collect();
                    debug!("reloaded certs");
                }
            }
        });
    }

    let app = Router::new()
        .typed_get(get_certs_identifier)
        .route("/certs/:identifier", post(post_certs_identifier))
        .route("/certs/:identifier", put(put_certs_identifier_update))
        .layer(ServiceBuilder::new().map_request_body(body::boxed))
        .with_state(state);

    // run our app with hyper
    // `axum::Server` is a re-export of `hyper::Server`
    tracing::debug!("listening on {}", address);
    axum::Server::bind(&address)
        .serve(app.into_make_service())
        .await
        .unwrap();
    Ok(())
}

pub enum ApiError {
    Internal,
    NotFound,
    Invalid,
    LowSerial(u64, u64),
}

type ApiResult<T> = Result<T, ApiError>;

impl IntoResponse for ApiError {
    fn into_response(self) -> axum::response::Response {
        match self {
            Self::NotFound => (StatusCode::NOT_FOUND, "not here").into_response(),
            Self::LowSerial(prev, next) => (
                StatusCode::BAD_REQUEST,
                format!(
                    "new certificate serial must be greater than {}, got {}",
                    prev, next
                ),
            )
                .into_response(),
            _ => (StatusCode::INTERNAL_SERVER_ERROR, "Oops").into_response(),
        }
    }
}

impl From<anyhow::Error> for ApiError {
    fn from(_: anyhow::Error) -> Self {
        ApiError::Internal
    }
}

#[derive(TypedPath, Deserialize)]
#[typed_path("/certs/:identifier")]
pub struct GetCert {
    pub identifier: String,
}

/// Retrieve an certificate for identifier
/// TODO: add option to require auth
/// return Unauthorized with an challenge
/// upon which the client will ssh-keysign
/// the challene an issue an post request
async fn get_certs_identifier(
    GetCert { identifier }: GetCert,
    State(ApiState { certs, .. }): State<ApiState>,
) -> ApiResult<String> {
    let certs = certs.lock().await;
    let cert = certs.get(&identifier).ok_or(ApiError::NotFound)?;
    Ok(cert.to_openssh().context("to openssh")?)
}

/// POST with signed challenge
async fn post_certs_identifier(
    State(ApiState { .. }): State<ApiState>,
    Path(_identifier): Path<String>,
) -> ApiResult<String> {
    unimplemented!()
}

/// Upload an cert with an higher serial than the previous
async fn put_certs_identifier_update(
    State(ApiState {
        ca,
        cert_dir,
        certs,
        ..
    }): State<ApiState>,
    Path(identifier): Path<String>,
    CertificateBody(cert): CertificateBody,
) -> ApiResult<String> {
    cert.validate(&[ca.fingerprint(Default::default())])
        .map_err(|_| ApiError::Invalid)?;
    let _string_repr = cert.to_openssh();
    let prev = load_cert(&cert_dir, &ca, &identifier).await?;
    let mut prev_serial = 0;
    let serial = cert.serial();
    if let Some(prev) = prev {
        prev_serial = prev.serial();
        if prev.serial() >= cert.serial() {
            return Err(ApiError::LowSerial(prev_serial, serial));
        }
    }
    store_cert(&cert_dir, &ca, &cert).await?;
    certs.lock().await.insert(cert.key_id().to_string(), cert);
    Ok(format!("{} -> {}", prev_serial, serial))
}
