use axum_extra::routing::TypedPath;
use clap::{Args, Parser, Subcommand};
use reqwest::StatusCode;
use ssh_key::Certificate;
use std::path::PathBuf;
use tokio::fs;
use tracing::{debug, info};
use url::Url;

use crate::{api::GetCert, certs::read_dir};

#[derive(Parser)]
pub struct ClientArgs {
    /// Url for the API endpoint
    #[clap(short = 'a', long = "api-endpoint")]
    api: Url,
}

#[derive(Parser)]
pub struct FetchArgs {
    #[clap(flatten)]
    args: ClientArgs,
    #[clap(short = 'c', long = "cert-dir", default_value = "~/.ssh")]
    cert_dir: PathBuf,
    /// CA public key
    #[clap(long = "ca")]
    ca: PathBuf,
}

#[derive(Args)]
pub struct ClientCommand {
    #[clap(subcommand)]
    cmd: ClientCommands,
}

#[derive(Subcommand)]
pub enum ClientCommands {
    Fetch(FetchArgs),
    Upload,
}

pub async fn run(ClientCommand { cmd }: ClientCommand) -> anyhow::Result<()> {
    match cmd {
        ClientCommands::Fetch(args) => fetch(args).await,
        ClientCommands::Upload => unimplemented!(),
    }
}

async fn fetch(
    FetchArgs {
        cert_dir,
        ca: _,
        args: ClientArgs { api },
    }: FetchArgs,
) -> anyhow::Result<()> {
    let certs = read_dir(&cert_dir).await?;
    let client = reqwest::Client::new();
    for cert in certs {
        let path = GetCert {
            identifier: cert.key_id().to_string(),
        };
        debug!("checking {}", cert.key_id());
        let url = api.join(path.to_uri().path())?;
        let resp = client.get(url.clone()).send().await?;
        if resp.status() != StatusCode::OK {
            continue;
        }
        let string_repr = resp.text().await?;
        let remote_cert = Certificate::from_openssh(&string_repr)?;
        if remote_cert
            .validate(&[cert.signature_key().fingerprint(Default::default())])
            .is_err()
        {
            info!("invalid signature {}, skipping", &url);
            continue;
        }
        if cert.serial() >= remote_cert.serial() {
            debug!("{} is not newer than local version", &url);
            continue;
        }
        fs::write(cert_dir.join(cert.key_id()), remote_cert.to_openssh()?).await?;
        info!(
            "updated {}: {} -> {}",
            cert.key_id(),
            cert.serial(),
            remote_cert.serial()
        );
    }
    Ok(())
}
